module m_poisson

  integer :: jacobi=0, gaussseidel=1

contains
  
  subroutine solve(T, U, F, nx, ny, method, maxiter, interval, maxerr)
    implicit none
    
    integer :: nx, ny, method
    real(kind=8), dimension(ny, nx) :: T, U, F
    real(kind=8) :: err, maxerr
    integer :: i, interval, maxiter
    
    if (.not.((method.eq.gaussseidel).or.(method.eq.jacobi))) then
       print*, 'Unknown method!'
       stop
    endif

!!$omp parallel shared(T, U, F, nx, ny)
    do i=1,maxiter ! we don't want to run *that* many iterations
       if (.false.) then
       !if (mod(i, interval).eq.0) then
       !   if (method.eq.gaussseidel) then
       !      call update_gs_and_calculate_norm(T, F, nx, ny, err)
       !   else
       !      call update_and_calculate_norm(T, U, F, nx, ny, err)
       !      T = U
       !   endif
       !   if(err.le.maxerr) then
       !      print*, 'Converged after', i, 'iterations'
       !      return
       !   endif
       else
          if (method.eq.gaussseidel) then
             call update_gs(T, F, nx, ny)
          elseif (method.eq.jacobi) then
             call update(T, U, F, nx, ny)
          endif
       endif
    enddo
!!$omp end parallel
    !print*, 'DID NOT CONVERGE IN MANY ITERATIONS'
  end subroutine solve

  subroutine update(T, U, F, nx, ny)
    implicit none

    integer, intent(in) :: nx, ny
    real(kind=8), dimension(ny, nx) :: T, U, F
        
    integer :: i, j
!$omp parallel
!$omp do private(i, j)
    do j=2, nx - 1
       do i=2, ny - 1
          U(i, j) = 0.25 * (T(i-1, j) + T(i+1, j) + T(i, j-1) + T(i, j+1) &
               + F(i, j))
       enddo
    enddo
!$omp end do

!$omp do private(i, j)
    do j=2, nx - 1
       do i=2, ny - 1
          T(i, j) = U(i, j)
       enddo
    enddo
!$omp end do
!$omp end parallel
  end subroutine update

  subroutine update_gs(T, F, nx, ny)
    implicit none

    integer :: nx, ny
    real(kind=8), dimension(ny, nx) :: T, F
        
    integer :: i, j

    do j=2, nx - 1
       do i=2, ny - 1
          T(i, j) = 0.25 * (T(i-1, j) + T(i+1, j) + T(i, j-1) + T(i, j+1) &
               + F(i, j))
       enddo
    enddo       

  end subroutine update_gs


  subroutine update_and_calculate_norm(T, U, F, nx, ny, norm)
    implicit none

    integer :: nx, ny
    real(kind=8), dimension(ny, nx) :: T, U, F
    real(kind=8) :: norm
    integer :: i, j
    real(kind=8) :: tmp
    norm = 0.0

    do j=2, nx - 1
       do i=2, ny - 1
          tmp = 0.25 * (T(i-1, j) + T(i+1, j) + T(i, j-1) + T(i, j+1) + F(i, j))
          norm = norm + (T(i, j) - tmp)**2
          U(i, j) = tmp
       enddo
    enddo
    norm = sqrt(norm)
    !call calculate_frobenius_norm(U, A, nx, ny, tmp)
    !print*, norm, tmp
  end subroutine update_and_calculate_norm

  subroutine update_gs_and_calculate_norm(T, F, nx, ny, norm)
    implicit none

    integer :: nx, ny
    real(kind=8), dimension(ny, nx) :: T, F
    real(kind=8) :: norm
    integer :: i, j
    real(kind=8) :: tmp
    norm = 0.0

    do j=2, nx - 1
       do i=2, ny - 1
          tmp = 0.25 * (T(i-1, j) + T(i+1, j) + T(i, j-1) + T(i, j+1) + F(i, j))
          norm = norm + (T(i, j) - tmp)**2
          T(i, j) = tmp
       enddo
    enddo
    norm = sqrt(norm)
    !call calculate_frobenius_norm(U, A, nx, ny, tmp)
    !print*, norm, tmp
  end subroutine update_gs_and_calculate_norm


  subroutine calculate_frobenius_norm(T, U, nx, ny, norm)
    implicit none
    
    integer :: nx, ny
    real(kind=8), dimension(ny, nx) :: T, U
    real(kind=8), intent(out) :: norm
    
    integer :: i, j

    real(kind=8) :: diff = 0.0
    norm = 0.0

    do j=1, nx
       do i=1, ny
          diff = T(i, j) - U(i, j)
          norm = norm + diff**2
       enddo
    enddo
    norm = sqrt(norm)
  end subroutine calculate_frobenius_norm

  subroutine initialize_boundary_conditions(T, U, F, nx, ny, dx, dy)
    implicit none
    
    integer :: nx, ny
    real(kind=8), dimension(ny, nx) :: T, U, F
    real(kind=8) :: dx, dy

    real(kind=8) :: temp = 20.0
    real(kind=8) :: radiatorfactor
    integer :: i, j
    integer :: xmin, xmax, ymin, ymax

    T = 0.0
    T(1, :) = temp ! 'north'
    T(:, 1) = temp ! 'east'
    T(ny, :) = temp ! 'south'
    T(:, nx) = 0.0 ! 'west'

    U = T

    F = 0.0
    
    xmin = 1 + nx / 2 ! account for one-indexation
    xmax = 1 + nx * 2 / 3
    ymin = 1 + ny / 6
    ymax = 1 + ny / 3

    radiatorfactor = 200.0 * dx * dy

    do i=xmin, xmax
       do j=ymin, ymax
          F(j, i) = radiatorfactor
       enddo
    enddo
    
  end subroutine initialize_boundary_conditions

  subroutine printarray(T, nx, ny)
    implicit none
    
    integer :: nx, ny
    real, dimension(nx, ny) :: T
    
    integer :: i
    
    do i=1, nx
       print*, T(i, :)
    enddo
  end subroutine printarray

  subroutine savearray(T, nx, ny, dx, dy, fileno)
    ! Save entire array to file as ASCII.
    implicit none
    
    integer, intent(in) :: fileno
    integer, intent(in) :: nx, ny
    real(kind=8), intent(in) :: dx, dy
    real(kind=8), dimension(ny, nx), intent(in) :: T

    integer :: i, j

    do j=1, nx
       do i=1, ny
          write(fileno, '(3E12.4)') real(i - 1) * dx, &
               real(j - 1) * dy, T(i, j)
       enddo
       write(fileno, '(A)')
    enddo
  end subroutine savearray
  

end module m_poisson
